#include <iostream>
#include <llvm/Passes/PassBuilder.h>
#include <llvm/IR/LLVMContext.h>
#include <llvm/IR/IRBuilder.h>
#include <llvm/IR/Verifier.h>
#include <llvm/IR/PassManager.h>
#include "llvm/Transforms/Scalar.h"
#include "llvm/Transforms/Scalar/Reassociate.h"
#include "llvm/Transforms/Scalar/GVN.h"
#include <llvm/Transforms/Scalar/DCE.h>
#include <llvm/Transforms/Scalar/SimplifyCFG.h>
#include <llvm/Transforms/InstCombine/InstCombine.h>
#include <llvm/Transforms/IPO/PassManagerBuilder.h>
#include <llvm/Analysis/MemorySSA.h>
#include <llvm/Analysis/PostDominators.h>
#include "ast.hpp"
#include "codegen.hpp"
#include "shared.h"

using namespace llvm;

static LLVMContext TheContext;
static IRBuilder<> Builder(TheContext);
static AnalysisManager<int, int> TheAM;
static std::map<std::string, Value *> NamedValues;
static std::unique_ptr<FunctionPassManager> TheFPM;
static std::unique_ptr<FunctionAnalysisManager> TheFAM;

void InitializeModuleAndPassManager(void) {
	TheModule = llvm::make_unique<Module>("Kaleidoscope jit", TheContext);
	
	TheFPM = make_unique<FunctionPassManager>();
	TheFAM = make_unique<FunctionAnalysisManager>();
	
	PassBuilder PB;
	PB.registerFunctionAnalyses(*TheFAM);
	
	TheFPM->addPass(InstCombinePass());
	TheFPM->addPass(ReassociatePass());
	TheFPM->addPass(SimplifyCFGPass());
	TheFPM->addPass(GVNHoistPass());
	TheFPM->addPass(GVNSinkPass());
}

std::map<std::string, std::unique_ptr<PrototypeAST>> functionProtos;

Function *getFunction(std::string name) {
	if (auto *func = TheModule->getFunction(name))
		return func;
	
	auto iterator = functionProtos.find(name);
	if (iterator != functionProtos.end())
		return iterator->second->codegen();
	
	return nullptr;
}

Value *LogErrorV(std::string str) {
	LogError(str);
	return nullptr;
}

Value *NumberExprAST::codegen() {
	return ConstantFP::get(TheContext, APFloat(Val));
}

Value *VariableExprAST::codegen() {
	Value *V = NamedValues[Name];
	if (!V)
		LogErrorV("Unknown variable name");
	return V;
}

Value *BinaryExprAST::codegen() {
	Value *L = LHS->codegen();
	Value *R = RHS->codegen();
	
	if (!L || !R) return nullptr;
	
	switch (Op) {
		case '+':
			return Builder.CreateFAdd(L, R, "addtmp");
		case '-':
			return Builder.CreateFSub(L, R, "subtmp");
		case '*':
			return Builder.CreateFMul(L, R, "multmp");
		case '<':
			L = Builder.CreateFCmpULT(L, R, "cmptmp");
			// convert bool 0/1 to double 0.0/1.0
			return Builder.CreateUIToFP(L, Type::getDoubleTy(TheContext), "booltmp");
		case '>':
			L = Builder.CreateFCmpUGT(L, R, "cmptmp");
			return Builder.CreateUIToFP(L, Type::getDoubleTy(TheContext), "booltmp");
		case '|':
			L = Builder.CreateFCmpUGE(L, ConstantFP::get(TheContext, APFloat(1.0)), "leftbooltmp");
			R = Builder.CreateFCmpUGE(R, ConstantFP::get(TheContext, APFloat(1.0)), "leftbooltmp");
			L = Builder.CreateOr(L, R, "orbooltmp");
			return Builder.CreateUIToFP(L, Type::getDoubleTy(TheContext), "ortmp");
		default:
			break;
	}

	Function *F = getFunction(std::string("binary") + Op);
	assert(F && "binary oprator not found!");

	Value *Ops[2] = {L, R};
	return Builder.CreateCall(F, Ops, "binop");
}

Value *CallExprAST::codegen() {
	Function *CalleeF = getFunction(Callee);
	if (!CalleeF)
		return LogErrorV("Unknown function referenced");
	
	if (CalleeF->arg_size() != Args.size())
		return LogErrorV("Incorrect number of arguments passed");
	
	std::vector<Value *> ArgsV;
	for (unsigned long i = 0, e = Args.size(); i != e; ++i) {
		ArgsV.push_back(Args[i]->codegen());
		if (!ArgsV.back()) return nullptr;
	}
	
	return Builder.CreateCall(CalleeF, ArgsV, "calltmp");
}

Function *PrototypeAST::codegen() {
	std::vector<Type*> Doubles(Args.size(), Type::getDoubleTy(TheContext));
	
	FunctionType *FT = FunctionType::get(Type::getDoubleTy(TheContext), Doubles, false);
	Function *func = Function::Create(FT, Function::ExternalLinkage, Name, TheModule.get());
	
	unsigned i = 0;
	for (auto &arg : func->args())
		arg.setName(Args[i++]);
	
	return func;
}

Function *FunctionAST::codegen() {
	//Transfer ownership but keep a reference
	auto &P = *Prototype;
	functionProtos[Prototype->getName()] = std::move(Prototype);
	
	Function *func = getFunction(P.getName());
	
	if (!func) return nullptr;

	if (P.isBinaryOp())
		BinopPrecedence[P.getOperatorName()] = P.getBinaryPrecedence();
	
	BasicBlock *bb = BasicBlock::Create(TheContext, "entry", func);
	Builder.SetInsertPoint(bb);
	
	NamedValues.clear();
	for (auto &arg: func->args())
		NamedValues[arg.getName()] = &arg;
	
	if (Value *retVal = Body->codegen()) {
		Builder.CreateRet(retVal);
		
		if (verifyFunction(*func, &errs())) {
			func->print(errs());
			func->eraseFromParent();
			return nullptr;
		}
		
		TheFPM->run(*func, *TheFAM);
		return func;
	}
	
	func->eraseFromParent();
	return nullptr;
}

Value *IfExprAST::codegen() {
	Value *condV = Cond->codegen();
	if (!condV) return nullptr;
	
	//convert to bool
	condV = Builder.CreateFCmpONE(condV, ConstantFP::get(TheContext, APFloat(0.0)), "ifcond");
	
	Function *func = Builder.GetInsertBlock()->getParent();
	
	BasicBlock *thenBB = BasicBlock::Create(TheContext, "then", func);
	BasicBlock *elseBB = BasicBlock::Create(TheContext, "else", func);
	BasicBlock *mergeBB = BasicBlock::Create(TheContext, "merge", func);
	
	Builder.CreateCondBr(condV, thenBB, elseBB);
	
	Builder.SetInsertPoint(thenBB);
	
	Value *thenV = Then->codegen();
	if (!thenV) return nullptr;
	
	Builder.CreateBr(mergeBB);
	
	//codegen of then can change the current block, so get then again
	thenBB = Builder.GetInsertBlock();
	
	Builder.SetInsertPoint(elseBB);
	
	Value *elseV = Else->codegen();
	if (!elseV) return nullptr;
	
	Builder.CreateBr(mergeBB);
	
	//codegen of else can change the current block, so get else again
	elseBB = Builder.GetInsertBlock();
	
	Builder.SetInsertPoint(mergeBB);
	PHINode *phiNode = Builder.CreatePHI(Type::getDoubleTy(TheContext), 2, "iftmp");
	phiNode->addIncoming(thenV, thenBB);
	phiNode->addIncoming(elseV, elseBB);
	
	return phiNode;
}

Value *ForExprAST::codegen() {
	auto *startValue = Start->codegen();
	if (!startValue) return nullptr;
	
	auto *func = Builder.GetInsertBlock()->getParent();
	auto *preheaderBB = Builder.GetInsertBlock();
	auto *loopBB = BasicBlock::Create(TheContext, "loop", func);
	
	Builder.CreateBr(loopBB);
	
	Builder.SetInsertPoint(loopBB);
	
	auto *index = Builder.CreatePHI(Type::getDoubleTy(TheContext), 2, VarName.c_str());
	index->addIncoming(startValue, preheaderBB);
	
	// if the index variable shadows an existing value, save it to restore later
	auto *oldVal = NamedValues[VarName];
	NamedValues[VarName] = index;
	
	// emit the loop body
	if (!Body->codegen()) return nullptr;
	
	Value *stepVal = nullptr;
	if (Step) {
		stepVal = Step->codegen();
		if (!stepVal) return nullptr;
	} else {
		stepVal = ConstantFP::get(TheContext, APFloat(1.0));
	}
	
	// increment the index
	auto *nextVar = Builder.CreateFAdd(index, stepVal, "nextvar");
	
	auto *endCond = End->codegen();
	if (!endCond) return nullptr;
	
	endCond = Builder.CreateFCmpONE(endCond, ConstantFP::get(TheContext, APFloat(0.0)), "loopcond");
	
	auto *loopEndBB = Builder.GetInsertBlock();
	auto *afterBB = BasicBlock::Create(TheContext, "afterloop", func);
	
	Builder.CreateCondBr(endCond, loopBB, afterBB);
	
	Builder.SetInsertPoint(afterBB);
	
	index->addIncoming(nextVar, loopEndBB);
	
	// restore shadowed index variable
	if (oldVal)
		NamedValues[VarName] = oldVal;
	else
		NamedValues.erase(VarName);
	
	return Constant::getNullValue(Type::getDoubleTy(TheContext));
}
