#include <cstdio>
#include <cctype>
#include <cstdlib>
#include <string>
#include <map>
#include <iostream>
#include <iomanip>
#include <llvm/ADT/STLExtras.h>
#include <llvm/Support/raw_ostream.h>
#include <llvm/ExecutionEngine/ExecutionEngine.h>
#include <llvm/ExecutionEngine/GenericValue.h>
#include <llvm/Support/TargetSelect.h>
//Needed to force linking interpreter
#include <llvm/ExecutionEngine/MCJIT.h>
#include "lexer.h"
#include "ast.hpp"
#include "codegen.hpp"
#include "shared.h"

static std::unique_ptr<ExprAST> ParseBinOpRHS(int ExprPrec,
											  std::unique_ptr<ExprAST> LHS);
static std::unique_ptr<ExprAST> ParseExpression();

static ExecutionEngine* TheEngine;

std::unique_ptr<Module> TheModule;

constexpr auto ANON_EXPR_FUNC_NAME = "__anon_expr";

enum Token {
	tok_eof = -1,
	
	tok_def = -2,
	tok_extern = -3,
	
	tok_identifier = -4,
	tok_number = -5,
	
	//control flow
	tok_if = -6,
	tok_then = -7,
	tok_else = -8,
	tok_for = -9,
	tok_in = -10,
	
	//operators
	tok_binary = -11,
	tok_unary = -12
};

static std::string IdentifierStr;
static double NumVal;

///Returns the next token from stdin
static int gettok() {
	static int LastChar = ' ';

	while (isspace(LastChar))
		LastChar = getchar();

	if (isalpha(LastChar)) {
		IdentifierStr = LastChar;
		while (isalnum((LastChar = getchar())))
			IdentifierStr += LastChar;

		if (IdentifierStr == "def") return tok_def;
		if (IdentifierStr == "extern") return tok_extern;
		if (IdentifierStr == "if") return tok_if;
		if (IdentifierStr == "then") return tok_then;
		if (IdentifierStr == "else") return tok_else;
		if (IdentifierStr == "binary") return tok_binary;
		if (IdentifierStr == "unary") return tok_unary;
		
		return tok_identifier;
	}

	if (isdigit(LastChar) || LastChar == '.') {
		std::string NumStr;
		do {
			NumStr += LastChar;
			LastChar = getchar();
		} while (isdigit(LastChar) || LastChar == '.');

		NumVal = strtod(NumStr.c_str(), 0);
		return tok_number;
	}

	if (LastChar == '#') {
		//Coment until the end of the line
		do
			LastChar = getchar();
		while(LastChar != EOF && LastChar != '\n' && LastChar != '\r');

		if(LastChar != EOF)
			return gettok();
	}

	//Check for end of file
	if (LastChar == EOF)
		return tok_eof;

	int ThisChar = LastChar;
	LastChar = getchar();
	return ThisChar;
}


static int CurTok;
static int getNextToken() {
	return CurTok = gettok();
}

std::unique_ptr<PrototypeAST> LogErrorP(std::string str) {
	LogError(str);
	return nullptr;
}

static std::unique_ptr<ExprAST> ParseNumberExpr() {
	auto Result = llvm::make_unique<NumberExprAST>(NumVal);
	getNextToken(); // consume the number
	return std::move(Result);
}

static std::unique_ptr<ExprAST> ParseParenExpr() {
	getNextToken();
	auto V = ParseExpression();
	if (!V)
		return nullptr;
	
	if (CurTok != ')')
		return LogError("expected ')'");
	getNextToken();
	return V;
}

static std::unique_ptr<ExprAST> ParseIdentifierExpr() {
	std::string IdName = IdentifierStr;
	
	getNextToken();
	
	if (CurTok != '(')
		return llvm::make_unique<VariableExprAST>(IdName);
	
	// call a function
	getNextToken();
	std::vector<std::unique_ptr<ExprAST>> Args;
	if (CurTok != ')') {
		while (true) {
			if (auto Arg = ParseExpression())
				Args.push_back(std::move(Arg));
			else
				return nullptr;
			
			if (CurTok == ')')
				break;
			
			if (CurTok != ',')
				return LogError("Expected ')' or ',' in argument list");
			getNextToken();
		}
	}
	
	getNextToken();
	return llvm::make_unique<CallExprAST>(IdName, std::move(Args));
}

static std::unique_ptr<ExprAST> ParseIfExpr() {
	getNextToken(); // eat the if
	
	auto condExpr = ParseExpression();
	if (!condExpr) return nullptr;
	
	if (CurTok != tok_then)
		return LogError("Expected then");
	getNextToken(); // eat the then
	
	auto thenExpr = ParseExpression();
	if (!thenExpr) return nullptr;
	
	if (CurTok != tok_else)
		return LogError("Expected else");
	getNextToken(); // eat the else
	
	auto elseExpr = ParseExpression();
	if (!elseExpr) return nullptr;
	
	return llvm::make_unique<IfExprAST>(std::move(condExpr),
										std::move(thenExpr),
										std::move(elseExpr));
}

static std::unique_ptr<ExprAST> ParseForExpr() {
	getNextToken(); // eat the for
	
	if (CurTok != tok_identifier)
		return LogError("Expected identifier after for");
	
	std::string id = IdentifierStr;
	getNextToken(); // eat the identifier
	
	if (CurTok != '=')
		return LogError("Expected = after for");
	getNextToken(); // eat the =
	
	auto start = ParseExpression();
	if (!start) return nullptr;
	
	if (CurTok != ',')
		return LogError("Expected , after for start value");
	getNextToken(); // eat the ,
	
	auto end = ParseExpression();
	if (!end)
		return nullptr;
	
	// optional step value
	std::unique_ptr<ExprAST> step;
	if (CurTok == ',') {
		getNextToken(); // eat the ,
		step = ParseExpression();
		if (!step) return nullptr;
	}
	
	if (CurTok != tok_in)
		return LogError("Expected in after for");
	getNextToken(); // eat the in
	
	auto body = ParseExpression();
	if (!body) return nullptr;
	
	return llvm::make_unique<ForExprAST>(id,
										 std::move(start),
										 std::move(end),
										 std::move(step),
										 std::move(body));
}

static std::unique_ptr<ExprAST> ParsePrimary() {
	switch (CurTok) {
		default:
			return LogError(std::string("Unknown token when expected an expression: ")
							+ std::to_string(CurTok));
		case tok_identifier:
			return ParseIdentifierExpr();
		case tok_number:
			return ParseNumberExpr();
		case '(':
			return ParseParenExpr();
		case tok_if:
			return ParseIfExpr();
		case tok_for:
			return ParseForExpr();
	}
}

static int GetTokPrecedence() {
	if (!isascii(CurTok))
		return -1;
	
	int TokPrec = BinopPrecedence[CurTok];
	if (TokPrec <= 0) return -1;
	return TokPrec;
}

static std::unique_ptr<ExprAST> ParseExpression() {
	auto LHS = ParsePrimary();
	if (!LHS)
		return nullptr;
	return ParseBinOpRHS(0, std::move(LHS));
}

static std::unique_ptr<ExprAST> ParseBinOpRHS(int ExprPrec,
											  std::unique_ptr<ExprAST> LHS) {
	while (true) {
		int TokPrec = GetTokPrecedence();
		if (TokPrec < ExprPrec)
			return LHS;
		
		int BinOp = CurTok;
		getNextToken(); // eat binop
		
		auto RHS = ParsePrimary();
		if (!RHS)
			return nullptr;
		
		int NextPrec = GetTokPrecedence();
		if (TokPrec < NextPrec) {
			RHS = ParseBinOpRHS(TokPrec + 1, std::move(RHS));
			if (!RHS)
				return nullptr;
		}
		
		LHS = llvm::make_unique<BinaryExprAST>(BinOp, std::move(LHS), std::move(RHS));
	}
}

static std::unique_ptr<PrototypeAST> ParsePrototype() {
	std::string FnName;
	
 	unsigned Kind = 0; // 0 = identifier, 1 = unary, 2 = binary
	unsigned BinaryPrecedence = 30;

	switch (CurTok) {
	default:
		return LogErrorP("Expected function name in prototype");
	case tok_identifier:
		FnName = IdentifierStr;
		Kind = 0;
		getNextToken();
		break;
	case tok_binary:
		getNextToken();
		if (!isascii(CurTok)) return LogErrorP("Expected binary operator");
		FnName = "binary";
		FnName += (char)CurTok;
		Kind = 2;
		getNextToken();
		if (CurTok == tok_number) {
			if (NumVal < 1 || NumVal > 100)
				return LogErrorP("Invalid precedence: must be 1..100");
			BinaryPrecedence = (unsigned)NumVal;
			getNextToken();
		}
		break;
	}
	
	if (CurTok != '(')
		return LogErrorP("Expected '(' in prototype");
	
	std::vector<std::string> ArgNames;
	while (getNextToken() == tok_identifier)
		ArgNames.push_back(IdentifierStr);
	if (CurTok != ')')
		return LogErrorP("Expected ')' in prototype");
	
	getNextToken(); // eat ')'

	if (Kind && ArgNames.size() != Kind)
		return LogErrorP("Invalid number of of arguments for operator");
	
	return llvm::make_unique<PrototypeAST>(FnName,
										   std::move(ArgNames),
										   Kind != 0,
										   BinaryPrecedence);
}

static std::unique_ptr<FunctionAST> ParseDefinition() {
	getNextToken();
	auto Proto = ParsePrototype();
	if (!Proto) return nullptr;
	
	if (auto E = ParseExpression())
		return llvm::make_unique<FunctionAST>(std::move(Proto), std::move(E));
	return nullptr;
}

static std::unique_ptr<PrototypeAST> ParseExtern() {
	getNextToken();
	return ParsePrototype();
}

static std::unique_ptr<FunctionAST> ParseTopLevelExpr() {
	if (auto e = ParseExpression()) {
		// make an anonymous prototype
		auto proto = llvm::make_unique<PrototypeAST>(ANON_EXPR_FUNC_NAME, std::vector<std::string>());
		return llvm::make_unique<FunctionAST>(std::move(proto), std::move(e));
	}
	return nullptr;
}

static void HandleDefinition() {
	if (auto def = ParseDefinition()) {
		if (auto gen = def->codegen()) {
			std::cerr << "Read function defintion:";
			gen->print(errs());
			std::cerr << "\n";
			
			TheEngine->addModule(std::move(TheModule));
			InitializeModuleAndPassManager();
		}
	}
	else
		getNextToken();	// skip token for error recovery
}

static void HandleExtern() {
	if (auto externProto = ParseExtern()) {
		if (auto gen = externProto->codegen()) {
			std::cerr << "Read an extern:\n";
			gen->print(errs());
			std::cerr << "\n";
			functionProtos[externProto->getName()] = std::move(externProto);
		}
	}
	else
		getNextToken(); // skip token for error recovery
}

static void HandleTopLevelExpr() {
	if (auto expr = ParseTopLevelExpr()) {
		if (auto gen = expr->codegen()) {
			auto module = TheModule.get();
			TheEngine->addModule(std::move(TheModule));
			InitializeModuleAndPassManager();
			
			auto func = TheEngine->FindFunctionNamed(ANON_EXPR_FUNC_NAME);
			GenericValue gv = TheEngine->runFunction(func, std::vector<GenericValue>());
			std::cerr << "Evaluated to " << std::fixed << std::setw(5) << gv.DoubleVal << std::endl;
			
			TheEngine->removeModule(module);
		}
	}
	else
		getNextToken(); // skip token for error recovery
}

void mainLoop() {
	while (true) {
		fprintf(stderr, "ready> ");
		switch(CurTok) {
			case tok_eof:
				return;
			case ';':
				getNextToken();
				break;
			case tok_def:
				HandleDefinition();
				break;
			case tok_extern:
				HandleExtern();
				break;
			default:
				HandleTopLevelExpr();
				break;
		}
	}
}

int main() {
	InitializeNativeTarget();
	InitializeNativeTargetAsmPrinter();
	InitializeNativeTargetAsmParser();
	
	BinopPrecedence['<'] = 10;
	BinopPrecedence['>'] = 10;
	BinopPrecedence['|'] = 5;
	BinopPrecedence['+'] = 20;
	BinopPrecedence['-'] = 20;
	BinopPrecedence['*'] = 40;
	
	InitializeModuleAndPassManager();
	
	std::string engineError;
	TheEngine = EngineBuilder(std::move(TheModule)).setEngineKind(EngineKind::JIT).setErrorStr(&engineError).create();
	if (!engineError.empty())
		std::cout << engineError << "\n";
	
	InitializeModuleAndPassManager();
	
	// prime the first token
	std::cerr << "ready> ";
	getNextToken();
	
	mainLoop();

	
	return 0;
}
